@@ -15,15 +15,16 @@ from buildbot.process.properties import WithProperties
 from buildbot.process.properties import Property
 from buildbot.schedulers.basic import SingleBranchScheduler
 
+from . import capability
 from scheduler import MirrorChangeFilter
 from utils import comma_list_sanitize
 from version import Version
 from version import VersionFilter
 
-BUILDSLAVE_KWARGS = { # name -> validating callable
-    'max_builds' : int,
+BUILDSLAVE_KWARGS = {  # name -> validating callable
+    'max_builds': int,
     'notify_on_missing': str,
-    }
+}
 
 BUILDSLAVE_REQUIRED = ('password',)
 
@@ -35,25 +36,49 @@ logger = logging.getLogger(__name__)
 # if they used shared eggs or downloads area
 buildout_lock = locks.SlaveLock("buildout")
 
+
 class BuildoutsConfigurator(object):
     """Populate buildmaster configs from buildouts and external slaves.cfg.
 
-    There are two dictionnaries of methods:
+    There are three dictionnaries of methods:
+       buildout_dl_steps : name -> method returning the main buildout config
+                           file name and a list of steps to
+                           construct the buildout configuration slave-side.
+       post_dl_steps: name-> method returning the main buildout config file
+                             name and a list of steps to be inserted
+                             between the buildout conf retrieval and the main
+                             buildout run.
        post_buildout_steps : name -> method returning a list of steps to be
                              added after buildout has run and testing db is
                              done
-       buildout_dl_steps : name -> method returning a list of steps to
-                           construct the buildout configuration slave-side.
     """
 
+    cap2environ = dict(
+        postgresql=dict(version_prop='pg_version',
+                        environ={'PGPORT': '%(cap(port):-)s',
+                                 'PGHOST': '%(cap(host):-)s',
+                                 'LD_LIBRARY_PATH': '%(cap(lib):-)s',
+                                 'PATH': '%(cap(bin):-)s',
+                                 'PGCLUSTER': '%(pg_version:-)s/main',
+                                 },
+                        ))
+
     def __init__(self, buildmaster_dir,
-                 manifest_paths=('buildouts/MANIFEST.cfg',)):
+                 manifest_paths=('buildouts/MANIFEST.cfg',),
+                 capability_options_to_environ=None):
         """Attach to buildmaster in which master_cfg_file path sits.
         """
         self.buildmaster_dir = buildmaster_dir
-        self.build_factories = {} # build factories by name
-        self.factories_to_builders = {} # factory name -> builders playing it
+        self.build_factories = {}  # build factories by name
+        self.factories_to_builders = {}  # factory name -> builders playing it
         self.manifest_paths = manifest_paths
+        if capability_options_to_environ is not None:
+            self.cap2environ = capability_options_to_environ
+
+    def add_capability_environ(self, capability_name, options2environ):
+        """Add a dict of capability options to environment mapping."""
+        self.cap2environ = self.cap2environ.copy()
+        self.cap2environ[capability_name] = options2environ
 
     def populate(self, config):
         config.setdefault('slaves', []).extend(self.make_slaves('slaves.cfg'))
@@ -91,27 +116,14 @@ class BuildoutsConfigurator(object):
         for slavename in parser.sections():
             kw = {}
             kw['properties'] = props = {}
-            props['capability'] = caps = {} # name -> versions
+            props['capability'] = caps = {}  # name -> versions
             seen = set()
             for key, value in parser.items(slavename):
                 seen.add(key)
                 if key in ('passwd', 'password'):
                     pwd = value
                 elif key == 'capability':
-                    for cap_line in value.split(os.linesep):
-                        split = cap_line.split()
-                        if len(split) == 1:
-                            name = split[0]
-                            version = None
-                        else:
-                            name = split[0]
-                            version = split[1]
-                        this_cap = caps.setdefault(name, {})
-                        cap_opts = this_cap.setdefault(version, {})
-                        for option in split[2:]:
-                            opt_name, opt_val = option.split('=')
-                            cap_opts[opt_name] = opt_val
-
+                    caps.update(capability.parse_slave_declaration(value))
                 elif key in BUILDSLAVE_KWARGS:
                     kw[key] = BUILDSLAVE_KWARGS[key](value)
                 else:
@@ -175,8 +187,36 @@ class BuildoutsConfigurator(object):
                 command=['python', 'buildout_hg_dl.py', url, branch],
                 description=("Retrieve buildout", "from hg",),
                 haltOnFailure=True,
-                )
             )
+        )
+
+    def buildout_hg_tag_dl_steps(self, cfg_tokens, manifest_dir):
+        """Steps to retrieve the buildout dir as a Mercurial tag.
+
+        Useful for release/packaging oriented builds.
+        The tag name is read from build properties.
+        The clone is made outside of the main build/ directory, that must
+        stay pristine to test the produced packages.
+        """
+
+        if len(cfg_tokens) != 2:
+            raise ValueError(
+                "Wrong hgtag buildout specification: %r" % cfg_tokens)
+
+        url, conf_path = cfg_tokens
+        tag = Property('buildout-tag')
+        return conf_path, (
+            FileDownload(
+                mastersrc=os.path.join(BUILD_UTILS_PATH, 'buildout_hg_dl.py'),
+                slavedest='../src/buildout_hg_dl.py',
+                haltOnFailure=True),
+            ShellCommand(
+                command=['python', 'buildout_hg_dl.py', '-t', 'tag', url, tag],
+                workdir='./src',
+                description=("Retrieve buildout", "tag", tag, "from hg",),
+                haltOnFailure=True,
+            )
+        )
 
     def make_factory(self, name, buildout_slave_path, buildout_dl_steps,
                      options):
@@ -192,6 +232,20 @@ class BuildoutsConfigurator(object):
         options is the config part for this factory, seen as a dict
         """
         factory = BuildFactory()
+        build_for = options.get('build-for')
+        factory.build_for = {}
+        if build_for is not None:
+            for line in build_for.split(os.linesep):
+                vf = VersionFilter.parse(line)
+                factory.build_for[vf.cap] = vf
+
+        requires = options.get('build-requires')
+        if requires is None:
+            factory.build_requires = []
+        else:
+            factory.build_requires = set(VersionFilter.parse(r)
+                                         for r in requires.split(os.linesep))
+
         factory.addStep(ShellCommand(command=['bzr', 'init-repo', '../..'],
                                      name="bzr repo",
                                      description="init bzr repo",
@@ -203,16 +257,26 @@ class BuildoutsConfigurator(object):
         for dl_step in buildout_dl_steps:
             factory.addStep(dl_step)
 
+        capability_env = capability.set_properties_make_environ(
+            self.cap2environ, factory)
+
+        for line in options.get('post-dl-steps', 'standard').split(os.linesep):
+            post_dl_steps = self.post_dl_steps[line]
+            buildout_slave_path, steps = post_dl_steps(
+                self, options, buildout_slave_path, environ=capability_env)
+            for step in steps:
+                factory.addStep(step)
+
         factory.addStep(FileDownload(
-                mastersrc=os.path.join(
-                    BUILD_UTILS_PATH, 'analyze_oerp_tests.py'),
-                slavedest='analyze_oerp_tests.py'))
+            mastersrc=os.path.join(
+                BUILD_UTILS_PATH, 'analyze_oerp_tests.py'),
+            slavedest='analyze_oerp_tests.py'))
+
         factory.addStep(PgSetProperties(
-            description=["Setting", "PG cluster", "properties"],
-            descriptionDone=["Set", "PG cluster", "properties"],
+            name, description=["Setting", "Testing DB", "property"],
+            descriptionDone=["Set", "Testing DB", "property"],
             name="pg_cluster_props",
-            factory_name=name,
-            ))
+        ))
 
         cache = '../../buildout-caches'
         eggs_cache = cache + '/eggs'
@@ -227,125 +291,103 @@ class BuildoutsConfigurator(object):
                                      haltOnFailure=True,
                                      ))
 
-        psycopg2_env=dict(PATH=[WithProperties('%(pg_bin:-)s'),
-                                '${PATH}'],
-                          LD_LIBRARY_PATH=WithProperties('%(pg_lib:-)s'),
-                          )
         factory.addStep(ShellCommand(command=[
-                    'bin/buildout',
-                    '-c', buildout_slave_path,
-                    'buildout:eggs-directory=' + eggs_cache,
-                    'buildout:openerp-downloads-directory=' + openerp_cache,
-                    'openerp:vcs-clear-locks=True',
-                    'openerp:vcs-clear-retry=True',
-                    WithProperties(
-                        'openerp:options.db_port=%(pg_port:-5432)s'),
-                    WithProperties(
-                        'openerp:options.db_host=%('
-                        'pg_host:-False)s'),
-                    WithProperties(
-                        'openerp:options.db_user=%(pg_user:-False)s'),
-                    WithProperties(
-                        'openerp:options.db_password=%(pg_passwd:-False)s'),
-                    ],
-                                     name="buildout",
-                                     description="buildout",
-                                     timeout=3600*4,
-                                     haltOnFailure=True,
-                                     locks=[buildout_lock.access('exclusive')],
-                                     env=psycopg2_env,
-                                     ))
-
-        # psql command and its environmental variables
-        psql = Property('pg_psql', default='psql')
-        psql_env = dict(PGHOST=WithProperties('%(pg_host:-)s'),
-                        PGPORT=WithProperties('%(pg_port:-)s'),
-                        PATH=[WithProperties('%(pg_bin:-)s'),
-                              '${PATH}'],
-                        )
+            'bin/buildout',
+            '-c', buildout_slave_path,
+            'buildout:eggs-directory=' + eggs_cache,
+            'buildout:openerp-downloads-directory=' + openerp_cache,
+            'openerp:with_devtools=True',
+            'openerp:vcs-clear-locks=True',
+            'openerp:vcs-clear-retry=True',
+            WithProperties(
+                'openerp:options.db_port=%(cap_postgresql_port:-5432)s'),
+            WithProperties(
+                'openerp:options.db_host=%(cap_postgresql_host:-False)s'),
+            WithProperties(
+                'openerp:options.db_user=%(cap_postgresql_user:-False)s'),
+            WithProperties(
+                'openerp:options.db_password='
+                '%(cap_postgresql_passwd:-False)s'),
+        ],
+            name="buildout",
+            description="buildout",
+            timeout=3600*4,
+            haltOnFailure=True,
+            locks=[buildout_lock.access('exclusive')],
+            env=capability_env,
+        ))
 
         factory.addStep(ShellCommand(command=[
-                    psql, 'postgres', '-c',
-                    WithProperties('DROP DATABASE IF EXISTS "%(testing_db)s"'),
-                    ],
-                                     name='dropdb',
-                                     description=["dropdb",
-                                                  Property('testing_db')],
-                                     env=psql_env,
-                                     haltOnFailure=True))
+            'psql', 'postgres', '-c',
+            WithProperties('DROP DATABASE IF EXISTS "%(testing_db)s"'),
+        ],
+            name='dropdb',
+            description=["dropdb", Property('testing_db')],
+            env=capability_env,
+            haltOnFailure=True,
+        ))
+
         factory.addStep(ShellCommand(command=[
-                    psql, 'postgres', '-c',
-                    WithProperties(
-            'CREATE DATABASE "%%(testing_db)s" '
-            'TEMPLATE "%s"' % options.get('db_template', 'template1')),
-                    ],
-                                     name='createdb',
-                                     description=["createdb",
-                                                  Property('testing_db')],
-                                     env=psql_env,
-                                     haltOnFailure=True,
-                                     ))
+            'psql', 'postgres', '-c',
+            WithProperties('CREATE DATABASE "%%(testing_db)s" '
+                           'TEMPLATE "%s"' % options.get('db_template',
+                                                         'template1')),
+        ],
+            name='createdb',
+            description=["createdb", Property('testing_db')],
+            env=capability_env,
+            haltOnFailure=True,
+        ))
 
-        post_buildout_steps = self.post_buildout_steps[
-            options.get('post-buildout-steps', 'standard')]
-        for step in post_buildout_steps(self, options, buildout_slave_path,
-                                        psycopg2_env=psycopg2_env):
-            factory.addStep(step)
+        for line in options.get('post-buildout-steps',
+                                'standard').split(os.linesep):
+            post_buildout_steps = self.post_buildout_steps[line]
 
-        # TODO GR this is outside of the factory itself and get back
-        # to the caller
-        build_for = options.get('build-for')
-        factory.build_for = {}
-        if build_for is not None:
-            for line in build_for.split(os.linesep):
-                vf = VersionFilter.parse(line)
-                factory.build_for[vf.cap] = vf
-
-        requires = options.get('build-requires')
-        if requires is None:
-            factory.build_requires = []
-        else:
-            factory.build_requires = [r.strip()
-                                      for r in requires.split(os.linesep)]
+            for step in post_buildout_steps(self, options, buildout_slave_path,
+                                            environ=capability_env):
+                factory.addStep(step)
 
         build_category = options.get('build-category')
         if build_category:
             factory.build_category = build_category.strip()
         return factory
 
-    def post_buildout_steps_standard(self, options, buildout_slave_path,
-                                     psycopg2_env=None):
+    def post_dl_steps_standard(self, options, buildout_slave_path, environ=()):
+        return buildout_slave_path, ()
 
-        if psycopg2_env is None:
-            psycopg2_env = {}
+    post_dl_steps = dict(standard=post_dl_steps_standard)
+
+    def post_buildout_steps_standard(self, options, buildout_slave_path,
+                                     environ=()):
+
+        environ = dict(environ)
 
         steps = []
 
         steps.append(ShellCommand(command=['rm', '-f', 'test.log'],
-                                     name="Log cleanup",
-                                     descriptionDone=['Cleaned', 'logs'],
-                                     ))
+                                  name="Log cleanup",
+                                  descriptionDone=['Cleaned', 'logs'],
+                                  ))
 
         steps.append(ShellCommand(command=[
-                    'bin/start_openerp', '-i',
-                    comma_list_sanitize(options.get('openerp-addons', 'all')),
-                    '--stop-after-init',
-                    '--log-level=test', '-d', Property('testing_db'),
-                    # openerp --logfile does not work with relative paths !
-                    WithProperties('--logfile=%(workdir)s/build/test.log')],
-                                     name='testing',
-                                     description='testing',
-                                     descriptionDone='tests',
-                                     logfiles=dict(test='test.log'),
-                                     haltOnFailure=True,
-                                     env=psycopg2_env,
-                                     )),
+            'bin/test_openerp', '-i',
+            comma_list_sanitize(options.get('openerp-addons', 'all')),
+            '-d', Property('testing_db'),
+            # openerp --logfile does not work with relative paths !
+            WithProperties('--logfile=%(workdir)s/build/test.log')],
+            name='testing',
+            description='testing',
+            descriptionDone='tests',
+            logfiles=dict(test='test.log'),
+            haltOnFailure=True,
+            env=environ,
+        ))
 
         steps.append(ShellCommand(
-                command=["python", "analyze_oerp_tests.py", "test.log"],
-                name='analyze',
-                description="analyze",
-                ))
+            command=["python", "analyze_oerp_tests.py", "test.log"],
+            name='analyze',
+            description="analyze",
+        ))
 
         return steps
 
@@ -378,15 +420,16 @@ class BuildoutsConfigurator(object):
             buildout_downloader = self.buildout_dl_steps.get(btype)
             if buildout_downloader is None:
                 raise ValueError("Buildout type %r in %r not supported" % (
-                        btype, name))
+                    btype, name))
 
             conf_slave_path, dl_steps = buildout_downloader(self, buildout[1:],
                                                             manifest_dir)
             registry[name] = factory = self.make_factory(
                 name, conf_slave_path, dl_steps, dict(parser.items(name)))
-            factory.manifest_path = manifest_path # change filter will need it
+            factory.manifest_path = manifest_path  # change filter will need it
 
     buildout_dl_steps = dict(standalone=buildout_standalone_dl_steps,
+                             hgtag=buildout_hg_tag_dl_steps,
                              hg=buildout_hg_dl_steps)
 
     def make_builders(self, master_config=None):
@@ -409,39 +452,54 @@ class BuildoutsConfigurator(object):
 
         slaves = master_config['slaves']
 
-        slaves_by_pg = {} # pg version -> list of slaves
+        slaves_by_pg = {}  # pg version -> list of slaves
         for slave in slaves:
-            for pg in slave.properties['capability'].get('postgresql', {}).keys():
+            for pg in slave.properties['capability'].get('postgresql', {}):
                 slaves_by_pg.setdefault(pg, []).append(slave)
 
         all_builders = []
         fact_to_builders = self.factories_to_builders
+
+        def only_if_requires(slave):
+            """Shorcut for extraction of build-only-if-requires tokens."""
+            only = slave.properties.getProperty('build-only-if-requires')
+            if only is None:
+                return set()
+            else:
+                return set(only.split())
+
         for factory_name, factory in self.build_factories.items():
             pgvf = factory.build_for.get('postgresql')
-            requires = set(factory.build_requires)
+            requires = factory.build_requires
+            require_names = set(req.cap for req in requires)
+            meet_requires = {}  # pg version -> list of slave names
 
-            meet_requires = {} # pg version -> list of slave names
             for pg, slaves in slaves_by_pg.items():
                 meet = [slave.slavename for slave in slaves
-                        if requires.issubset(slave.properties['capability'])]
+                        if capability.does_meet_requirements(
+                            slave.properties['capability'], requires)
+                        and only_if_requires(slave).issubset(require_names)
+                        ]
                 if meet:
                     meet_requires[pg] = meet
 
             builders = [
-                BuilderConfig(name='%s-postgresql-%s' % (factory_name,
-                                                         pg_version),
-                              properties=dict(pg_version=pg_version),
-                              category=getattr(factory, 'build_category', None),
-                              factory=factory, slavenames=slavenames)
+                BuilderConfig(
+                    name='%s-postgresql-%s' % (factory_name,
+                                               pg_version),
+                    properties=dict(pg_version=pg_version),
+                    category=getattr(factory, 'build_category', None),
+                    factory=factory, slavenames=slavenames)
                 for pg_version, slavenames in meet_requires.items()
                 if pgvf is None or pgvf.match(Version.parse(pg_version))
-                ]
+            ]
+
             fact_to_builders[factory_name] = [b.name for b in builders]
             all_builders.extend(builders)
         return all_builders
 
     def factory_to_manifest(self, fact_name, absolute=False):
-        """Return the path to manifest file where factory with given name arose.
+        """Return the path to manifest file where factory fact_name arose.
         """
         path = self.build_factories[fact_name].manifest_path
         if absolute:
@@ -456,6 +514,7 @@ class BuildoutsConfigurator(object):
         not be preferable for buildmaster performance.
         """
         fact_to_builders = self.factories_to_builders
+
         def make_filter(factory_name):
             """Make a Mirror Change Filter for factory with given name."""
             return MirrorChangeFilter(
